""" Handles requests and data un-entanglement to a truly retarded web service. """
from __future__ import annotations

from urllib.parse import urlencode
from logging import getLogger

import binascii
import aiohttp

from .constants import (
    NO_CONTROLS,
    NO_ACCESS_RESPONSES, API_URL, CLIENT_TIMEOUT, CLIENT_VERSION_NUMBER,
)
from .utils import (
    map_response_data, get_query_params, aes_encrypt, aes_decrypt,
)
from .models import (
    Account,
    AuthStatus,
)
from .exceptions import (
    NoAccessException,
    AuthenticationError,
)

_LOGGER = getLogger(__name__)


class Client:
    def __init__(self, user: Account | None = None):
        self.user = user

    @classmethod
    def initialize(cls, username: str, password: str):
        return cls(Account(**{username: username, password: password}))

    @classmethod
    def login(cls, username: str, password: str) -> Client:
        c = cls()
        c.authenticate_user(username, password)
        return c

    async def api_request(self, params: dict, headers: dict | None = None):
        method = params.get('p')
        if method != 'l' and self.user:
            params.update(self.user.get_query_params())
        data = await self.do_external_api_request(params, headers)
        if data in NO_ACCESS_RESPONSES:
            raise NoAccessException()
        _LOGGER.debug("Got response: %s", data)
        return data

    async def do_external_api_request(
        self,
        params: dict,
        headers: dict | None = None
    ):
        if headers is None:
            headers = {}

        payload = get_query_params(params)
        _LOGGER.debug("Doing API request with params: %s", payload)
        url = f'{API_URL}/app.php?{aes_encrypt(urlencode(payload))}'
        headers = {
            'user-agent': f'PK_{CLIENT_VERSION_NUMBER}',
            **headers,
        }

        async with aiohttp.ClientSession(
            raise_for_status=True,
            timeout=aiohttp.ClientTimeout(CLIENT_TIMEOUT),
        ) as session:
            async with session.get(url, headers=headers) as resp:
                enc_data = await resp.text('utf-8')
                try:
                    data = aes_decrypt(enc_data)
                except binascii.Error:
                    data = enc_data

                return data.strip('\x00\x01\x02\x03\x04\x05\x06\x07\x08\x10\x0f').strip()

    def set_user(self, user: Account):
        self.user = user

    async def authenticate_user(self, username: str, password: str):
        auth_user = Account(**{"username": username, "password": password})
        params = {
            'p': 'l',
            'lang': 'no',
            **auth_user.get_query_params(),
        }

        result = await self.api_request(params)
        _LOGGER.debug("Got result: %s", result)
        user_dict = map_response_data(result, [
            'status',
            'country',
            None,
            'phone_prefix',
            'state',
            'uid',
        ])
        if user_dict.get('status') == AuthStatus.LOGIN_ERROR:
            raise AuthenticationError

        account_dict = {
            **auth_user.dict(),
            **user_dict,
            **{
                "username": f"{user_dict['phone_prefix']}{auth_user.username}",
            }
        }

        account = Account(**{str(k): str(v) for k, v in account_dict.items()})
        self.set_user(account)
        return account

    async def get_settings(self):
        params = {
            'p': 'instillinger',
        }
        return await self.api_request(params)

    async def get_control(self, cid: int):
        params = {
            'p': 'hki',
            'kontroll_id': cid,
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'id',  # 14241
            'county',  # Trøndelag
            'municipality',  # Malvik
            'type',  # Fartskontroll
            'timestamp',  # 29.05 - 20:47
            'description',  # Kontroll Olderdalen
            'lat',  # 63.4258007013951
            'lng',  # 10.6856604194473
            None,  # |
            None,  # |
            None,  # malvik.png
            None,  # trondelag.png
            'speed_limit',  # 90
            None,  # 1
            'last_seen',  # 20:47
            'confirmed',  # 0
            None,  # 2
            None,  # 1
        ])

    async def get_controls(
        self,
        lat: float,
        lng: float,
        **kwargs,
    ):
        params = {
            'p': 'hk',
            **{
                'lat': lat,
                'lon': lng,
            },
            **kwargs,
        }

        result = await self.api_request(params)
        if result == NO_CONTROLS:
            return []

        return map_response_data(result, [
            'id',  # 14239
            'county',  # Trøndelag
            'municipality',  # Meråker
            'type',  # Toll/grense
            None,  # 20:02
            'description',  # Toll
            'lat',  # 63.3621679609569
            'lng',  # 11.9694197550416
            None,  # NOT_IN_USE
            None,  # meraaker.png
            None,  # YES
            None,  # meraaker.png
            'timestamp',  # 1685383334
            None,  # 0
            None,  # 20:04
            'last_seen',  # 1685383471
        ], multiple=True)

    async def get_controls_in_radius(
        self,
        lat: float,
        lng: float,
        radius: int,
        speed: int = 100,
        **kwargs,
    ):
        return await self.get_controls(lat,
                                       lng,
                                       p='gps_kontroller',
                                       vr=radius,
                                       speed=speed,
                                       **kwargs)

    async def get_control_types(self):
        params = {
            'p': 'kontrolltyper',
        }
        result = await self.api_request(params)

        data = map_response_data(result, [
            'slug',
            'name',
            'id',
            None,
        ], multiple=True)
        return [
            {
                # Remove ".png"
                key: val[:-4] if key == 'slug' else val
                for key, val in el.items()
            }
            for el in data
        ]

    async def get_maps(self):
        params = {
            'p': 'hent_mine_kart',
        }
        result = await self.api_request(params)
        data = map_response_data(result, [
            'id',
            None,
            'title',
            'country',
        ], multiple=True)
        return data

    async def exchange_points(self):
        params = {
            'p': 'veksle',
        }
        result = await self.api_request(params)
        data = map_response_data(result, [
            'status',
            'message',
        ])
        return data



""" Politikontroller models """
import re
from datetime import datetime
from enum import Enum
from pydantic import BaseModel, validator

from .utils import parse_time_format
from .constants import DEFAULT_COUNTRY, PHONE_PREFIXES


class AuthStatus(str, Enum):
    LOGIN_OK = 'LOGIN_OK'
    LOGIN_ERROR = 'LOGIN_ERROR'


class ExchangeStatus(str, Enum):
    EXCHANGE_OK = 'EXCHANGE_OK'


class PoliceControlTypeEnum(str, Enum):
    SPEED_TRAP = "Fartskontroll"
    BEHAVIOUR = "Belte/mobil"
    TECHNICAL = "Teknisk"
    TRAFFIC_INFO = "Trafikk info"
    TRAFFIC_MESSAGE = "Trafikkmelding"
    OBSERVATION = "Observasjon"
    CUSTOMS = "Toll/grense"
    WEIGHT = "Vektkontroll"
    UNKNOWN = "Ukjent"
    CIVIL_POLICE = "Sivilpoliti"
    MC_CONTROL = "Mopedkontroll"
    BOAT_PATROL = "Politibåten"


class Account(BaseModel):
    uid: int | None
    status: AuthStatus | None
    country: str = DEFAULT_COUNTRY
    username: str
    password: str | None
    state: str | None

    @property
    def phone_number(self):
        return int(self.username[2:]) if len(self.username) > 8 else int(self.username)

    @property
    def phone_prefix(self):
        return int(self.username[:2]) if len(self.username) > 8 \
            else PHONE_PREFIXES.get(self.country.lower())

    @validator('username', pre=True)
    def validate_username(cls, v):
        v = re.sub(r' ', '', str(v))
        return v

    def get_query_params(self):
        """ Get query params. """
        return {
            'retning': self.phone_prefix,
            'telefon': self.phone_number,
            'passord': self.password,
        }


class PoliceControlType(BaseModel):
    id: int
    name: PoliceControlTypeEnum
    slug: str


class PoliceControlPoint:
    type: str = "Point"
    lat: float
    lng: float

    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    @property
    def coordinates(self):
        return self.lng, self.lat

    @property
    def __geo_interface__(self):
        return {
            'type': self.type,
            'coordinates': self.coordinates,
        }


class PoliceControl(BaseModel):
    id: int
    type: PoliceControlTypeEnum
    county: str
    speed_limit: int | None = None
    municipality: str
    description: str
    lat: float
    lng: float
    timestamp: datetime | None
    last_seen: datetime | None
    confirmed: int = 0

    @validator('timestamp', pre=True)
    def timestamp_validate(cls, v):
        if len(v) == 0 or (v.isnumeric() and int(v) == 0):
            return None
        return parse_time_format(v)

    @validator('last_seen', pre=True)
    def last_seen_validate(cls, v):
        if len(v) == 0 or (v.isnumeric() and int(v) == 0):
            return None
        return parse_time_format(v)

    @property
    def description_truncated(self):
        return (
            self.description[:25] + '..'
        ) if len(self.description) > 27 else self.description

    @property
    def title(self):
        return f"{self.type.value}: {self.description_truncated}"

    @property
    def _geometry(self):
        return PoliceControlPoint(self.lat, self.lng)

    @property
    def __geo_interface__(self):
        return {
            "type": "Feature",
            "geometry": self._geometry.__geo_interface__,
            "properties": {
                "title": self.title,
                "description": self.description,
                "type": self.type,
            },
        }


class ExchangePointsResponse(BaseModel):
    status: ExchangeStatus
    message: str



from datetime import datetime
from logging import getLogger
from json import JSONDecoder, JSONEncoder
import random
import string
import time
import base64
import re
from Crypto.Cipher import AES

from .constants import (
    CRYPTO_K1,
    CRYPTO_K2,
    CLIENT_OS,
    CLIENT_VERSION_NUMBER,
)

_LOGGER = getLogger(__name__)


def get_random_string(length: int, letters: str | None = None) -> str:
    if letters is None:
        letters = string.ascii_uppercase + string.digits
    return ''.join(random.choice(letters) for _ in range(length))


def generate_device_id():
    return get_random_string(16, letters=string.digits + 'abcdef')


def get_unix_timestamp():
    return int(time.time()) + 10


def hash_credentials(credentials: dict):
    creds = bytes(JSONEncoder().encode(credentials), 'utf-8')
    return base64.b64encode(creds).decode()


def unhash_credentials(credentials: str):
    creds = base64.b64decode(credentials).decode()
    return JSONDecoder().decode(creds)


def aes_encrypt(input_str: str):
    """
    Encrypts a string using AES encryption with given key and initialization vector.
    Returns base64-encoded result.
    """
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    length = 16 - (len(input_str) % 16)
    input_b = bytes(input_str, 'utf-8') + bytes([length]) * length
    return base64.b64encode(cipher.encrypt(input_b)).decode()


def aes_decrypt(enc_base64: str):
    """
    Decrypts AES encrypted data using a given key and initialization vector.
    """
    enc_data = base64.b64decode(enc_base64)
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    decipher = AES.new(key, AES.MODE_CBC, iv)
    return decipher.decrypt(enc_data).decode()


def get_query_params(params: dict):
    """
    Generates a query dictionary with random and predefined values.
    Replaces special characters in the values with hyphens.
    """
    query = {
        'bac': get_random_string(10),
        'z': int(time.time()) + 10,
        'version': CLIENT_VERSION_NUMBER,
        'os': CLIENT_OS,
        **params,
        'tt': get_random_string(5),
    }
    for k, v in query.items():
        if k in ['bac', 'z', 'version', 'os', 'tt']:
            query[k] = re.sub(r"[|#\\\"]", "-", str(v))
    return query

def map_response_data(
    data: str,
    map_keys: list[str | None],
    multiple=False
) -> list[dict[str, str]] | dict[str, str]:
    """Converts a cvs-like string into dictionaries."""

    def row_to_dict(row) -> dict[str, str]:
        r = dict(zip(map_keys, row.split('|')))  # pylint: disable=modified-iterating-list
        return {k: r[k] for k in r.keys() if isinstance(k, str)}

    if multiple:
        return list(map(row_to_dict, list(data.split('#'))))

    return row_to_dict(data)


def parse_time_format(text: str):
    today = datetime.today()
    try:
        return int(
            datetime.strptime(text, "%d.%m - %H:%M").replace(
                year=today.year,
            ).timestamp()
        )
    except ValueError:
        pass

    try:
        return int(
            datetime.strptime(text, "%H:%M").replace(
                year=today.year,
                month=today.month,
                day=today.day,
            ).timestamp()
        )
    except ValueError:
        pass
    return text



CRYPTO_KEY_SIZE = 128
CRYPTO_K1 = "aGFua19lcl9ob21vX3Bvbw=="  # "hank_er_homo_poo"
CRYPTO_K2 = "ZGlsZG9zZXJ2aWNlX3N1eA=="  # "dildoservice_sux"
CLIENT_VERSION_NUMBER = '9.1.0'
CLIENT_OS = 'Android'
CLIENT_TIMEOUT = 30
API_URL = 'http://app.politikontroller.no'

NO_ACCESS_RESPONSES = [
    'USER_NOT_AUTHORIZED_NOPREM',
]
NO_CONTROLS = 'INGEN_KONTROLLER'

PHONE_PREFIXES = {
    "no": 47,
    "se": 46,
    "dk": 45,
}
DEFAULT_COUNTRY = "no"



""" CLI tool """

import logging
from tabulate import tabulate
from rich import print as rprint
import anyio
import asyncclick as click

from politikontroller_py import Client
from politikontroller_py.models import (
    PoliceControl,
)
from politikontroller_py.exceptions import AuthenticationError

TABULATE_DEFAULTS = {
    'tablefmt': 'rounded_grid',
}

client = Client()


@click.group()
@click.option('--username', '-u', envvar='POLITIKONTROLLER_USERNAME', type=str,
              required=True, help='Username (i.e. phone number)')
@click.password_option('--password', '-p', envvar='POLITIKONTROLLER_PASSWORD', type=str, required=True,
                       confirmation_prompt=False, help='Password')
@click.option('--debug', is_flag=True, help='Set logging level to DEBUG')
async def cli(username: str, password: str, debug: bool):
    """
    Username and password can be defined using env vars:

    POLITIKONTROLLER_USERNAME
    POLITIKONTROLLER_PASSWORD
    """
    configure_logging(debug)
    try:
        await client.authenticate_user(username, password)
    except AuthenticationError as err:
        raise click.BadParameter("Authentication error",
                                 param_hint=["--username", "--password"]) from err


@cli.command('get-controls', short_help='get a list of all active controls.')
@click.option('--lat', required=True, help='Your position (latitude)')
@click.option('--lng', required=True, help='Your position (longitude)')
async def get_controls(lat: float, lng: float):
    controls = await client.get_controls(lat, lng)
    click.echo(tabulate(controls, **TABULATE_DEFAULTS))


@cli.command('get-controls-radius',
             short_help='get all active controls inside a radius.')
@click.option('--lat', type=float, required=True, help='Radius center (latitude)')
@click.option('--lng', type=float, required=True, help='Radius center (longitude)')
@click.option('--radius', type=int, required=True, metavar='km',
              help='Radius size in kilometers')
@click.option('--speed', type=int, required=False, metavar='km/h',
              help='Speed, unknown what this does')
async def get_controls_in_radius(lat: float, lng: float, radius: int, speed: int):
    controls = await client.get_controls_in_radius(lat, lng, radius, speed)
    click.echo(tabulate(controls, **TABULATE_DEFAULTS))


@cli.command('get-control', short_help='get details on a control.')
@click.argument('control_id', type=int, required=True)
async def get_control(control_id: int):
    control = PoliceControl(**await client.get_control(control_id))
    rprint(control)


@cli.command('get-maps', short_help='get own maps.')
async def get_maps():
    maps = await client.get_maps()
    rprint(maps)


@cli.command('exchange-points', short_help='exchange points (?)')
async def exchange_points():
    res = await client.exchange_points()
    rprint(res)

def configure_logging(debug: bool = False):
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)

if __name__ == '__main__':
    anyio.run(cli)
